@@ -4,9 +4,6 @@
 #include <tut/tut.hpp>
 // geos
 #include <geos_c.h>
-// std
-#include <cstdarg>
-#include <cstdio>
 
 #include "capi_test_utils.h"
 
@@ -60,5 +57,18 @@ void object::test<2>
     GEOSGeom_destroy(expected);
 }
 
+
+template<>
+template<>
+void object::test<3>()
+{
+    input_ = fromWKT("MULTICURVE (CIRCULARSTRING (0 0, 1 1, 2 0), (2 0, 3 0))");
+    ensure(input_);
+
+    result_ = GEOSLineMergeDirected(input_);
+
+    ensure("curved geometries not supported", result_ == nullptr);
+}
+
 } // namesplace tut
 
